from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from askoclics.askolib.client import Client

from future import standard_library

standard_library.install_aliases()


class DatasetClient(Client):
    """
    Manipulate datasets managed by Askomics
    """

    def list(self):
        """
        List datasets added in Askomics

        :rtype: list
        :return: List of datasets
        """

        return self._api_call("get", "list_datasets", {})['datasets']

    def delete(self, datasets):
        """
        Send a delete task on a list of datasets

        :type datasets: str
        :param datasets: Comma-separated list of datasets IDs

        :rtype: list
        :return: List of the datasets
        """

        datasets = self._parse_input_values(datasets, "Datasets")
        body = {'datasetsIdToDelete': datasets}

        return self._api_call("post", "delete_datasets", body)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from future import standard_library

standard_library.install_aliases()


class AskoclicsApiError(Exception):
    """Raised when the API returns an error"""
    pass


class AskoclicsConnectionError(Exception):
    """Raised when the connection to the Askomics server fails"""
    pass


class AskoclicsAuthError(Exception):
    """Raised when the login with the provided api key failed"""
    pass


class AskoclicsParametersError(Exception):
    """Raised when parameters are missing"""
    pass


class AskoclicsNotImplementedError(Exception):
    """Raised when the endpoint does not exists"""
    pass



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from distutils.version import StrictVersion

from askoclics.askolib.dataset import DatasetClient
from askoclics.askolib.exceptions import AskoclicsAuthError, AskoclicsConnectionError, AskoclicsNotImplementedError
from askoclics.askolib.file import FileClient
from askoclics.askolib.result import ResultClient
from askoclics.askolib.sparql import SparqlClient

from future import standard_library

import requests

standard_library.install_aliases()


class AskomicsInstance(object):

    def __init__(self, api_key=None, url="http://localhost:80", proxy_username="", proxy_password="", **kwargs):

        if not api_key:
            raise AskoclicsAuthError("An api key is required")

        if proxy_username and proxy_password:
            self.auth = (proxy_username, proxy_password)
        else:
            self.auth = None

        self.api_key = api_key

        url = url.rstrip().rstrip("/")

        self.url = url

        self._check_connectivity()

        self.endpoints = self._get_endpoints()

        # Initialize Clients
        args = (self.url, self.endpoints, self.api_key, self.auth)
        self.file = FileClient(*args)
        self.dataset = DatasetClient(*args)
        self.result = ResultClient(*args)
        self.sparql = SparqlClient(*args)

    def __str__(self):
        return '<GopublishInstance at {}>'.format(self.url)

    def _check_connectivity(self):
        headers = {"X-API-KEY": self.api_key}

        try:
            r = requests.get("{}/api/start".format(self.url), headers=headers, auth=self.auth)
            if not r.status_code == 200:
                raise requests.exceptions.RequestException

            data = r.json()

            if StrictVersion(data['config']['version']) < StrictVersion('4.3.0'):
                raise AskoclicsNotImplementedError("Askomics server version is older than 4.3.0.")

            if not data['config'].get("logged"):
                raise AskoclicsAuthError("Could not login with the provided API key.")

        except requests.exceptions.RequestException:
            raise AskoclicsConnectionError("Cannot connect to {}. Please check the connection.".format(self.url))

    def _get_endpoints(self):

        endpoints = {
            "start": "/api/start",
            "sparql_init": "/api/sparql/init",
            "sparql_preview": "/api/sparql/previewquery",
            "sparql_query": "/api/sparql/savequery",
            "upload_local_file": "/api/files/upload_chunk",
            "upload_url_file": "/api/files/upload_url",
            "list_files": "/api/files",
            "preview_files": "/api/files/preview",
            "delete_files": "/api/files/delete",
            "integrate_file": "/api/files/integrate",
            "list_datasets": "/api/datasets",
            "publicize_dataset": "/api/datasets/public",
            "delete_datasets": "/api/datasets/delete",
            "list_results": "/api/results",
            "preview_results": "/api/results/preview",
            "download_results": "/api/results/download",
            "delete_results": "/api/results/delete",
            "sparql_results": "/api/results/sparqlquery"
        }
        return endpoints


__version__ = '0.0.1'

PROJECT_NAME = "askoclics"



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import mimetypes
import os

from askoclics.askolib.client import Client
from askoclics.askolib.exceptions import AskoclicsParametersError

from future import standard_library

import requests

standard_library.install_aliases()


class FileClient(Client):
    """
    Manipulate files managed by AskOmics
    """

    def list(self):
        """
        List files added in AskOmics

        :rtype: list
        :return: List with files
        """

        return self._api_call("get", "list_files", {})['files']

    def upload(self, url="", file_path="", verbose=False):
        """
        Upload a file to AskOmics

        :type url: str
        :param url: URL to the file

        :type file_path: str
        :param file_path: Path to the file to upload

        :type verbose: bool
        :param verbose: Show progression bar for local file upload

        :rtype: dict
        :return: Dict with results
        """
        if not (url or file_path) or (url and file_path):
            raise AskoclicsParametersError("Please provided either an url or a file path")

        if url:
            return self._api_call("post", "upload_url_file", {"url": url})

        if not os.path.isfile(file_path):
            raise AskoclicsParametersError("Local file not found")

        file_name = os.path.basename(file_path)
        internal_path = file_name
        mimetype = mimetypes.guess_type(file_path)[0]
        # Chunk size to 10 Mo
        file_size = os.stat(file_path).st_size
        chunk_size = 1024 * 1024 * 10
        first = True
        last = False
        uploaded_size = 0

        if file_size <= chunk_size:
            last = True

        with open(file_path, "rb") as f:
            if verbose:
                print("0%")

            for piece in self._read_in_chunks(f, chunk_size):
                if uploaded_size + chunk_size >= file_size:
                    last = True

                body = {"chunk": piece.decode("utf-8"), "first": first, "last": last, "type": mimetype, "name": file_name, "size": file_size, "path": internal_path}
                res = self._api_call("post", "upload_local_file", body)
                first = False
                internal_path = res["path"]
                uploaded_size += chunk_size
                if uploaded_size > file_size:
                    uploaded_size = file_size

                if verbose:
                    print("{0:.0%}".format(uploaded_size / file_size))

        return res

    def preview(self, files):
        """
        Get preview for a list of files

        :type files: str
        :param files: Comma-separated file IDs

        :rtype: dict
        :return: Dictionary containing the information
        """

        files = self._parse_input_values(files, "Files")
        body = {'filesId': files}

        return self._api_call("post", "preview_files", body)

    def describe(self, files):
        """
        Show file information

        :type files: str
        :param files: Comma-separated file IDs

        :rtype: list
        :return: List of files containing info
        """

        files = self._parse_input_values(files, "Files")
        body = {'filesId': files}

        res = self._api_call("post", "preview_files", body)

        files = []

        for file in res.get("previewFiles"):
            if "data" in file:
                file["data"].pop("content_preview", None)
            files.append(file)

        return files

    def integrate_csv(self, file_id, columns="", headers="", force=False, custom_uri=None, skip_preview=False):
        """
        Send an integration task for a specified file_id

        :type file_id: str
        :param file_id: File_id

        :type columns: str
        :param columns: Comma-separated columns (default to detected columns)

        :type headers: str
        :param headers: Comma-separated headers (default to file headers)

        :type force: bool
        :param force: Ignore the content type mismatch (ex: force an integer type when AskOmics detects a text type)

        :type custom_uri: str
        :param custom_uri: Custom uri

        :type skip_preview: bool
        :param skip_preview: Skip the preview step for big files

        :rtype: dict
        :return: Dictionary of task information
        """

        columns = self._parse_input_values(columns, "Columns")
        headers = self._parse_input_values(headers, "Headers")

        if not skip_preview:

            data = self._check_integrate_file(file_id, "csv/tsv")

            if headers and not len(headers) == len(data["header"]):
                raise AskoclicsParametersError("Incorrect number of headers : {} headers supplied, {} headers expected".format(len(headers), len(data["header"])))

            if columns:
                if not len(columns) == len(data["columns_type"]):
                    raise AskoclicsParametersError("Incorrect number of columns : {} columns supplied, {} columns expected".format(len(columns), len(data["columns_type"])))

                expected_columns = self._get_column_types()
                for index, val in enumerate(columns):
                    if index == 0:
                        if val not in ["start_entity", "entity"]:
                            raise AskoclicsParametersError("First column type must be either start_entity or entity")
                        continue

                    if val not in expected_columns:
                        raise AskoclicsParametersError("Column type {} is not supported by AskOmics. Supported column types are {}".format(val, expected_columns))

                if not force:
                    for index, value in enumerate(data["columns_type"]):
                        if value == "text" and columns[index] not in ["text", "category", "general_relation", "symetric_relation"]:
                            raise AskoclicsParametersError("Type mismatch on provided column {} : provided type is {}, but AskOmics predicted {}. To proceed, use the force parameter".format(index + 1, columns[index], value))

        body = {"fileId": file_id, "columns_type": columns, "header_names": headers, "customUri": custom_uri}
        return self._api_call("post", "integrate_file", body)

    def integrate_bed(self, file_id, entity="", custom_uri=None, skip_preview=False):
        """
        Send an integration task for a specified file_id

        :type file_id: str
        :param file_id: File_id

        :type entity: str
        :param entity: Name of the entity (default to file name)

        :type custom_uri: str
        :param custom_uri: Custom uri

        :type skip_preview: bool
        :param skip_preview: Skip the preview step for big files

        :rtype: dict
        :return: Dictionary of task information
        """

        if not skip_preview:
            self._check_integrate_file(file_id, "bed")

        body = {"fileId": file_id, "entity_name": entity, "customUri": custom_uri}
        return self._api_call("post", "integrate_file", body)

    def integrate_gff(self, file_id, entities="", custom_uri=None, skip_preview=False):
        """
        Send an integration task for a specified file_id

        :type file_id: str
        :param file_id: File_id

        :type entities: str
        :param entities: Comma-separated list of entities to integrate. (Default to all available entities)

        :type custom_uri: str
        :param custom_uri: Custom uri

        :type skip_preview: bool
        :param skip_preview: Skip the preview step for big files

        :rtype: dict
        :return: Dictionary of task information
        """

        entities = self._parse_input_values(entities, "Entities")

        if not skip_preview:
            data = self._check_integrate_file(file_id, "gff/gff3")
            if entities:
                for entity in entities:
                    if entity not in data['entities']:
                        AskoclicsParametersError("Entity {} was not detected in the file. Detected entities are : {} ".format(entity, data['entities']))

        body = {"fileId": file_id, "entities": entities, "customUri": custom_uri}
        return self._api_call("post", "integrate_file", body)

    def integrate_rdf(self, file_id, external_endpoint=None, skip_preview=False):
        """
        Send an integration task for a specified file_id

        :type file_id: str
        :param file_id: File_id

        :type external_endpoint: str
        :param external_endpoint: External endpoint

        :type skip_preview: bool
        :param skip_preview: Skip the preview step for big files

        :rtype: dict
        :return: Dictionary of task information

        """

        if not skip_preview:
            self._check_integrate_file(file_id, "rdf/ttl")

        body = {"fileId": file_id, "externalEndpoint": external_endpoint}
        return self._api_call("post", "integrate_file", body)

    def delete(self, files):
        """
        Delete a list of files

        :type files: str
        :param files: Comma-separated file IDs to delete

        :rtype: dict
        :return: Dictionary containing the remaining files
        """

        files = self._parse_input_values(files, "Files")
        body = {'filesIdToDelete': files}

        return self._api_call("post", "delete_files", body)

    def _read_in_chunks(self, file_object, chunk_size):
        """Lazy function (generator) to read a file piece by piece.
        Default chunk size: 10 Mo."""
        while True:
            data = file_object.read(chunk_size)
            if not data:
                break
            yield data

    def _check_integrate_file(self, file_id, file_type):
        # Check if file exists
        # Then check if type is correct
        # Return predicted columns/types if csv/bed

        body = {'filesId': [file_id]}
        res = self._api_call("post", "preview_files", body)

        files = res.get("previewFiles")

        if not files:
            raise AskoclicsParametersError("File with id {} not found".format(file_id))

        file = files[0]

        if file["error"]:
            raise AskoclicsParametersError("File is in error state : {}".format(file["error_message"]))

        if not file["type"] == file_type:
            raise AskoclicsParametersError("Incorrect file type : Selected type is {}, selected file type is {}".format(file_type, file["type"]))

        return file.get("data")

    def _get_column_types(self):
        r = requests.get("{}/api/files/columns".format(self.url), auth=self.auth)
        if not r.status_code == 200:
            raise requests.exceptions.RequestException

        data = r.json()
        return data["types"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os

from askoclics.askolib.client import Client
from askoclics.askolib.exceptions import AskoclicsParametersError

from future import standard_library


standard_library.install_aliases()


class ResultClient(Client):
    """
    Interact with AskOmics results
    """

    def list(self):
        """
        List results

        :rtype: dict
        :return: Dict with info
        """
        data = self._api_call("get", "list_results", {})
        data.pop('triplestoreMaxRows', None)
        return data

    def preview(self, result_id):
        """
        Show results preview

        :type result_id: str
        :param result_id: Result id

        :rtype: dict
        :return: Dict with info
        """
        return self._api_call("post", "preview_results", {"fileId": result_id})

    def download(self, result_id, file_path):
        """
        Download a result to a file

        :type result_id: str
        :param result_id: Result id

        :type file_path: str
        :param file_path: Output file path

        :rtype: None
        :return: None
        """

        if os.path.exists(file_path):
            raise AskoclicsParametersError("File {} already exists".format(file_path))
        data = self._api_call("post", "download_results", {"fileId": result_id}, download=True)
        with open(file_path, "w") as f:
            f.write(data)

    def delete(self, result_ids):
        """
        Delete results

        :type result_ids: str
        :param result_ids: Comma-separated result ids

        :rtype: None
        :return: None
        """
        data = {}
        data["filesIdToDelete"] = self._parse_input_values(result_ids, "Result ids")
        return self._api_call("post", "delete_results", data)

    def get_sparql(self, result_id):
        """
        Show results preview

        :type result_id: str
        :param result_id: Result id

        :rtype: dict
        :return: Dict with info
        """

        data = self._api_call("post", "sparql_results", {"fileId": result_id})
        data.pop("console_enabled", None)
        data.pop("disk_space", None)
        return data



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json
import re

from askoclics.askolib.exceptions import AskoclicsApiError, AskoclicsConnectionError, AskoclicsNotImplementedError, AskoclicsParametersError

from future import standard_library

import requests

standard_library.install_aliases()


class Client(object):
    """
    Base client class implementing methods to make queries to the server
    """

    def __init__(self, url, endpoints, api_key, auth):
        self.url = url
        self.endpoints = endpoints
        self.api_key = api_key
        self.auth = auth

    def _api_call(self, call_type, endpoint_name, body={}, inline=False, download=False):

        headers = {"X-API-KEY": self.api_key}
        url = self._format_url(call_type, endpoint_name, body)

        try:
            if call_type == "get":
                if inline:
                    r = requests.get(url, params=body, headers=headers, auth=self.auth)
                else:
                    r = requests.get(url, headers=headers, auth=self.auth)
            elif call_type == "post":
                r = requests.post(url, json=body, headers=headers, auth=self.auth)

            if 400 <= r.status_code <= 499:
                mes = "API call returned a {} error".format(r.status_code)
                try:
                    res = r.json()
                    if res.get('errorMessage'):
                        mes += " : {}".format(res.get('errorMessage'))
                except json.decoder.JSONDecodeError:
                    pass
                raise AskoclicsApiError(mes)
            elif r.status_code == 502:
                raise AskoclicsApiError("Unknown server error")
            if download:
                return r.content
            res = r.json()
            if res.get('error'):
                raise AskoclicsApiError(res.get('errorMessage'))
            return res

        except requests.exceptions.RequestException:
            raise AskoclicsConnectionError("Cannot connect to {}. Please check the connection.".format(self.url))

    def _format_url(self, call_type, endpoint_name, body, inline=False):

        endpoint = self.endpoints.get(endpoint_name)
        if not endpoint:
            raise AskoclicsNotImplementedError()

        # Fill parameters in the url
        if not inline and call_type in ["get", "delete"]:
            groups = re.findall(r'<(.*?)>', endpoint)
            for group in groups:
                if group not in body:
                    raise AskoclicsApiError("Missing get parameter " + group)
                endpoint = endpoint.replace("<{}>".format(group), body.get(group))

        return "{}{}".format(self.url, endpoint)

    def _parse_input_values(self, val, val_name):

        if not val:
            return []

        if isinstance(val, list):
            return val
        elif isinstance(val, str):
            return [data.strip() for data in val.split(",")]
        else:
            raise AskoclicsParametersError("{} must either be a list or a comma-separated string".format(val_name))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os

from askoclics.askolib.client import Client

from future import standard_library


standard_library.install_aliases()


class SparqlClient(Client):
    """
    Send SPARQL queries to Askomics
    """

    def info(self):
        """
        Return available graphs, endpoints, and uris

        :rtype: dict
        :return: Dict with info
        """
        data = {}
        config = self._api_call("get", "start", {})
        data["namespace_data"] = config["config"]["namespaceData"]
        data["namespace_internal"] = config["config"]["namespaceInternal"]
        sparql_config = self._api_call("get", "sparql_init", {})
        data["graphs"] = sparql_config["graphs"]
        data["endpoints"] = sparql_config["endpoints"]
        return data

    def template(self, file_path):
        """
        Write the default query to a file

        :type file_path: str
        :param file_path: Path to the file to create

        :rtype: None
        :return: None
        """
        sparql_config = self._api_call("get", "sparql_init", {})
        with open(file_path, "w") as f:
            f.write(sparql_config["defaultQuery"])

    def query(self, query, graphs, endpoints, full_query=False):
        """
        Send a SPARQL query

        :type query: str
        :param query: Either a path to a file, or the query as a string

        :type graphs: str
        :param graphs: Comma-separated graphs

        :type endpoints: str
        :param endpoints: Comma-separated endpoints

        :type full_query: bool
        :param full_query: Whether to send a full query or a preview

        :rtype: dict
        :return: The API call result (either the result id, or the result preview)
        """
        data = {}

        if os.path.isfile(query):
            with open(query, 'r') as f:
                query = f.read()

        data["query"] = query
        data["graphs"] = self._parse_input_values(graphs, "Graphs")
        data["endpoints"] = self._parse_input_values(endpoints, "Endpoints")

        if full_query:
            return self._api_call("post", "sparql_query", data)
        else:
            return self._api_call("post", "sparql_preview", data)
